from lokii.config import CONFIG

from lokii.storage.data_storage import DataStorage


class BatchIterator:
    """
    :var data_storage: database storage client
    :type data_storage: DataStorage
    :var node_name: name of the node
    :type node_name: str
    :var cur_batch: current batch index
    :type cur_batch: int
    """

    def __init__(self, data_storage, node_name):
        """
        :param data_storage: database storage client
        :type data_storage: DataStorage
        :param node_name: name of the node
        :type node_name: str
        """
        self.data_storage = data_storage
        self.node_name = node_name
        self.cur_batch = 0

    def __iter__(self):
        return self

    def __next__(self):
        q = "SELECT * FROM %s" % self.node_name
        data = self.data_storage.exec(q, self.cur_batch, CONFIG.gen.batch_size)
        if len(data) > 0:
            self.cur_batch += 1
            return data
        raise StopIteration



import os
import pytest
from lokii.config import CONFIG

from lokii.parse.node_parser import NodeParser

pytestmark = [pytest.mark.usefixtures("glob_files", "load_modules")]


@pytest.mark.parametrize("glob_files, load_modules", [([], [])], indirect=True)
def test_parse_should_log_warning_if_no_modules_found(caplog):
    NodeParser("/test/path").parse()
    assert "No generation node file found" in caplog.text


@pytest.mark.parametrize(
    "glob_files, load_modules",
    [
        (
            ["/test/path/t1.node.py", "/test/path/g1/t2.node.py"],
            [
                {"source": "SELECT 1", "item": lambda x: x},
                {"source": "SELECT 1", "item": lambda x: x},
            ],
        )
    ],
    indirect=True,
)
def test_parse_should_use_filename_if_name_not_provided(glob_files, load_modules):
    parsed = NodeParser("/test/path").parse()
    for module_path in glob_files:
        node_name = os.path.basename(module_path).replace(CONFIG.gen.node_ext, "")
        assert node_name == parsed[node_name].name


@pytest.mark.parametrize(
    "glob_files, load_modules",
    [
        (
            ["/test/path/t1.node.py", "/test/path/g1/t2.node.py"],
            [
                {"source": "SELECT 1", "item": lambda x: x, "name": "test3"},
                {"source": "SELECT 1", "item": lambda x: x, "name": "test4"},
            ],
        )
    ],
    indirect=True,
)
def test_parse_should_use_name_instead_of_filename(glob_files, load_modules):
    parsed = NodeParser("/test/path").parse()
    for i, module_path in enumerate(glob_files):
        node_name = load_modules[i]["name"]
        assert node_name == parsed[node_name].name


@pytest.mark.parametrize(
    "glob_files, load_modules, expect",
    [
        (
            ["/test/path/g1/t1.node.py", "/test/path/g1/g2/t2.node.py"],
            [
                {"source": "SELECT 1", "item": lambda x: x, "name": "test3"},
                {"source": "SELECT 1", "item": lambda x: x, "name": "test4"},
            ],
            [["g1"], ["g1", "g2"]],
        )
    ],
    indirect=["glob_files", "load_modules"],
)
def test_parse_should_use_file_path_to_extract_groups(glob_files, load_modules, expect):
    parser = NodeParser("/test/path")
    parser.parse()
    for i, module_path in enumerate(glob_files):
        node_name = load_modules[i]["name"]
        assert expect[i] == parser.nodes[node_name].groups


@pytest.mark.parametrize("glob_files", [["/test/path/g1/t1.node.py"]], indirect=True)
@pytest.mark.parametrize(
    "load_modules, expect",
    [
        ([{}], "`source` not found"),
        ([{"source": 1000}], "`source` must be str"),
        ([{"source": "1"}], "`item` not found"),
        ([{"source": "1", "item": ""}], "`item` must be function"),
        ([{"source": "1", "item": lambda x, y: x}], "`item` accepts only one param"),
        ([{"source": "1", "item": lambda x: x, "name": 10}], "`name` must be str"),
    ],
    indirect=["load_modules"],
)
def test_parse_raise_error_if_module_not_valid(expect):
    with pytest.raises(AssertionError) as err:
        NodeParser("/test/path").parse()
    assert expect in str(err.value)



from collections import defaultdict


class GraphAnalyzer:
    def __init__(self, nodes: list[(str, list[str])]):
        """
        Reads and validates dataset configuration from filesystem structure.

        :param nodes: root path of the dataset generation schema
        """
        self.graph = defaultdict(list)
        self.nodes = []
        for node, deps in nodes:
            self.nodes.append(node)
            self.graph[node].extend(deps)

    def __cycle(self, index, node, visited, stack):
        visited[index] = True
        stack[index] = True

        for neighbour in self.graph[node]:
            n_index = self.nodes.index(neighbour)
            if not visited[n_index]:
                if self.__cycle(n_index, neighbour, visited, stack):
                    return True
            elif stack[n_index]:
                return True
        stack[index] = False
        return False

    def check_cyclic(self):
        length = len(self.nodes) + 1
        visited = [False] * length
        stack = [False] * length
        for index, node in enumerate(self.nodes):
            if not visited[index] and self.__cycle(index, node, visited, stack):
                raise AssertionError("Found cyclic dependencies!\n%s" % node)

    def __sort(self, index, node, visited, stack):
        visited[index] = True
        for element in self.graph[node]:
            e_index = self.nodes.index(element)
            if not visited[e_index]:
                self.__sort(e_index, element, visited, stack)
        stack.insert(0, node)

    def topological_sort(self):
        stack = []
        visited = [False] * len(self.nodes)
        for index, node in enumerate(self.nodes):
            if not visited[index]:
                self.__sort(index, node, visited, stack)
        return stack[::-1]

    def dependencies(self, name: str) -> list[str]:
        stack = []
        visited = [False] * len(self.nodes)
        index = self.nodes.index(name)
        self.__sort(index, name, visited, stack)
        return stack[::-1]

    def execution_order(self) -> list[str]:
        self.check_cyclic()
        return self.topological_sort()



import pytest

from util.graph_analyzer import GraphAnalyzer


def test_dependencies_should_return_correct_run_ids():
    n1, n2, n3, n4 = (("n1", ["n3"]), ("n2", ["n1", "n3"]), ("n3", []), ("n4", ["n2"]))
    analyzer = GraphAnalyzer([n1, n2, n3, n4])
    deps = analyzer.dependencies(n2[0])
    assert deps == [n3[0], n1[0], n2[0]]


def test_execution_order_should_raise_error_if_cyclic_dependencies_exists():
    n1, n2, n3 = (("n1", ["n3"]), ("n2", ["n1"]), ("n3", ["n2"]))
    analyzer = GraphAnalyzer([n1, n2, n3])
    with pytest.raises(AssertionError) as err:
        analyzer.execution_order()
    assert "Found cyclic dependencies" in str(err.value)


def test_execution_order_should_return_run_ids_in_correct_order():
    n1, n2, n3, n4 = (("n1", ["n3"]), ("n2", ["n1", "n3"]), ("n3", []), ("n4", ["n2"]))
    analyzer = GraphAnalyzer([n1, n2, n3, n4])
    order = analyzer.execution_order()
    assert order == [n3[0], n1[0], n2[0], n4[0]]



import argparse
import logging
import sys

from lokii import Lokii
from lokii.logger.context import LoggingContext
from lokii.config import CONFIG

LOKII_ASCII = r"""
▄▄▄▄▄   ▄▄▄▄▄▄▄▄▄▄▄▄▄ ▄▄▄▄▄▄▄▄▄▄▄
█   █   █       █   █ █ █   █   █
█   █   █   ▄   █   █▄█ █   █   █
█   █   █  █ █  █      ▄█   █   █
█   █▄▄▄█  █▄█  █     █▄█   █   █
█       █       █    ▄  █   █   █
█▄▄▄▄▄▄▄█▄▄▄▄▄▄▄█▄▄▄█ █▄█▄▄▄█▄▄▄█ 
"""


class Command:
    def __init__(self, argv=None) -> None:
        self.argv = argv.split() if argv else sys.argv[:]
        self.prog_name = "lokii"

    def execute(self) -> None:
        """
        Given the command-line arguments, this creates a parser appropriate
        to that command, and runs it.
        """

        formatter_class = argparse.RawDescriptionHelpFormatter
        parser = argparse.ArgumentParser(
            prog=self.prog_name,
            description="%s\n%s version %s"
            % (LOKII_ASCII, self.prog_name, CONFIG.version),
            formatter_class=formatter_class,
        )

        parser.add_argument(
            "--version", action="version", version="%(prog)s " + CONFIG.version
        )

        parser.add_argument(
            "-v",
            "--verbose",
            action="store_true",
            help="set logging level to DEBUG that shows all events",
        )

        parser.add_argument(
            "-q",
            "--quiet",
            action="store_true",
            help="set logging level to CRITICAL that shows only blocking events",
        )

        parser.add_argument(
            "-l",
            "--log-file",
            action="store",
            help="set the file path where the log file will be created",
        )

        parser.add_argument(
            "-f",
            "--source-folder",
            action="store",
            metavar="PATH",
            default=".",
            type=str,
            help="generate the specified number of outputs",
        )

        parser.add_argument(
            "-e",
            "--export",
            action="store_true",
            help="execute group export modules",
        )

        parser.add_argument(
            "-p",
            "--purge",
            action="store_true",
            help="purge cache database after generation",
        )

        arguments = parser.parse_args(self.argv[1:])

        verbose = logging.INFO
        if arguments.verbose:
            verbose = logging.DEBUG
        if arguments.quiet:
            verbose = logging.ERROR

        with LoggingContext(level=verbose, filename=arguments.log_file):
            try:
                print(LOKII_ASCII)
                _lokii = Lokii(arguments.source_folder)
                _lokii.generate(arguments.export, arguments.purge)
            except Exception as err:
                logging.critical(str(err), exc_info=True)


def exec_cmd(argv=None) -> None:
    command = Command(argv)
    command.execute()



import os
import shutil

import pytest
from pytest import FixtureRequest
from pytest_mock import MockerFixture


from lokii.config import CONFIG
from lokii import Lokii


@pytest.fixture
def glob_files(mocker, request):
    """
    :param mocker: Fixture that provides the same interface to functions in the mock module,
    ensuring that they are uninstalled at the end of each test.
    :type mocker: MockerFixture
    :param request: A request for a fixture from a test or fixture function.
    :type request: FixtureRequest
    :return: list of found files
    :rtype: list[str]
    """
    files = request.param if hasattr(request, "param") else []
    files = [os.path.normpath(f) for f in files]
    mocker.patch(
        "lokii.parse.group_parser.glob",
        return_value=[f for f in files if CONFIG.gen.group_ext in f],
    )
    mocker.patch(
        "lokii.parse.node_parser.glob",
        return_value=[f for f in files if CONFIG.gen.node_ext in f],
    )
    return files


@pytest.fixture
def load_modules(mocker, request, glob_files):
    """
    :param mocker: Fixture that provides the same interface to functions in the mock module,
    ensuring that they are uninstalled at the end of each test.
    :type mocker: MockerFixture
    :param request: A request for a fixture from a test or fixture function.
    :type request: FixtureRequest
    :param glob_files: A glob mock fixture from a test or fixture function.
    :type glob_files: list[str]
    """

    def module_file_loader_side_effect(file_path: str):
        mods = request.param if hasattr(request, "param") else []
        mod_i = glob_files.index(file_path)
        loader = {
            "load": mocker.Mock(),
            "module": type("object", (object,), mods[mod_i]),
            "version": mods[mod_i]["version"] if "version" in mods[mod_i] else "v1",
        }
        return type("ModuleFileLoader", (object,), loader or {})()

    group_mock = mocker.patch("lokii.parse.group_parser.ModuleFileLoader")
    group_mock.side_effect = module_file_loader_side_effect

    node_mock = mocker.patch("lokii.parse.node_parser.ModuleFileLoader")
    node_mock.side_effect = module_file_loader_side_effect
    return request.param if hasattr(request, "param") else []


@pytest.fixture(scope="function")
def setup_test_env(request):
    def teardown():
        Lokii.clean_env(force=True)
        if os.path.exists("data"):
            shutil.rmtree("data")

    Lokii.setup_env()
    request.addfinalizer(teardown)



import inspect
from typing import Any


class BaseParser:
    def attr(self, obj: object, name: str, msg: str = None):
        return self.__test(hasattr(obj, name), msg)

    def inst(self, obj: object, info: Any, msg: str = None):
        return self.__test(isinstance(obj, info), msg)

    def func(self, obj: object, msg: str = None):
        return self.__test(inspect.isfunction(obj), msg)

    def sig(self, obj: Any, args_count: int, msg: str = None):
        parameter_count = len(inspect.signature(obj).parameters)
        return self.__test(parameter_count == args_count, msg)

    def __test(self, cond: bool, msg: str = None):
        if not msg:
            return cond
        assert cond, msg



import time


class PerfTimerContext:
    """
    Simple perf_counter implementation that tracks elapsed time in context.
    """

    def __enter__(self):
        self.time = time.perf_counter()
        return self

    def __exit__(self, _, __, ___):
        self.time = time.perf_counter() - self.time

    def __str__(self):
        units = [(3600, "h"), (60, "m"), (1, "s")]
        try:
            measure, unit = next((self.time / s, u) for s, u in units if self.time > s)
            return "{:.2f}{}".format(measure, unit)
        except StopIteration:
            return "{:.4f}s".format(self.time)



import math
from functools import partial
from typing import Callable

from pathos.pools import ProcessPool

from lokii.config import CONFIG
from lokii.model.node_module import GenNodeModule
from lokii.logger.progress import ProgressLogger
from lokii.storage.data_storage import DataStorage
from lokii.storage.temp_storage import TempStorage


def _exec_chunk(func: Callable, args: list[dict]):
    return [func(arg) for arg in args]


class NodeExecutor:
    def __init__(self, node: GenNodeModule, data_storage: DataStorage):
        """
        Reads and validates dataset configuration from filesystem structure.

        :param node: root path of the dataset generation
        :param data_storage: root path of the dataset generation
        """
        self.run = node
        self.data_storage = data_storage
        self.__temp_storage = TempStorage(self.run.name)

        # total times the gen function will be called
        self.t_count = 0
        # total generated item count
        self.g_count = 0

    def prepare_node(self) -> int:
        self.t_count = self.data_storage.count(self.run.source)
        return self.t_count

    def exec_node(self) -> list[str]:
        logger = ProgressLogger(self.t_count)
        batch_size = CONFIG.gen.batch_size
        batch_count = math.ceil(self.t_count / batch_size)

        for batch_index in range(batch_count):
            # calculate indexes from `f` to `t`
            f = batch_index * batch_size
            t = self.t_count if f + batch_size > self.t_count else f + batch_size

            params = self.data_storage.exec(self.run.source, batch_index, batch_size)
            args = [{"index": i, "id": i + 1, "params": params[i]} for i in range(f, t)]

            batch_result = []
            for chunk_result in self._exec_batch(args):
                batch_result.extend(chunk_result)
                logger.update(len(chunk_result))

            # save result to temp storage
            self.__temp_storage.dump(batch_result)
            self.g_count += len(batch_result)

        # return generated file paths
        return self.__temp_storage.batches

    def _exec_batch(self, args: list[dict]) -> list[dict]:
        chunk_size = CONFIG.gen.chunk_size
        chunk_args = [args[i : i + chunk_size] for i in range(0, len(args), chunk_size)]

        gen_func = partial(_exec_chunk, self.run.item)
        with ProcessPool(nodes=CONFIG.gen.concurrency) as pool:
            for chunk in pool.uimap(gen_func, chunk_args):
                # remove null items from generated chunk
                yield [i for i in chunk if i is not None]



import os
import sys
import hashlib
import inspect
from types import ModuleType


class ModuleFileLoader:
    """
    :type module: ModuleType or None
    :type version: str or None
    """

    def __init__(self, file_path: str):
        """
        Loads module from given path.

        :param file_path: file path of the module
        """
        self.path = os.path.abspath(file_path)
        self.filename = os.path.basename(file_path)

        self.module = None
        self.version = None

    def load(self) -> None:
        if not os.path.exists(self.path):
            raise FileNotFoundError("No module file found at: %s" % self.path)

        try:
            import importlib.util

            # read file spec from given path. filename is used as module name
            spec = importlib.util.spec_from_file_location(self.filename, self.path)
            module = importlib.util.module_from_spec(spec)

            # introduce module to system
            sys.modules[self.filename] = module
            spec.loader.exec_module(module)
            self.module = module
        except ImportError:
            from imp import load_source

            module = load_source(self.filename, self.path)

        # acquire module source code as str
        src = inspect.getsource(module).encode("utf-8")
        h = hashlib.blake2b(digest_size=20)
        h.update(src)

        # hash source code to detect changes
        self.version = h.hexdigest()



import logging

from lokii.logger.formatter import MultiFormatter


class LoggingContext:
    """
    Use a logging configuration for a CLI application.
    This will prettify log messages for the console, and show more info in a log file.
    :param logger: the logger to configure. If None, configures the root logger
    :param level:  sets the output verbosity
    :param filename:  the file name of the log file to log to. If None, no log file is generated.
    :return: a context manager that will configure the logger, and reset to the previous
    configuration afterward.
    """

    def __init__(
        self,
        logger: logging.Logger = None,
        level: int = logging.INFO,
        filename: str = None,
    ):
        self.logger = logger or logging.root
        self.level = level
        self.old_level = self.logger.level

        console_handler = logging.StreamHandler()
        console_handler.setFormatter(MultiFormatter())
        console_handler.setLevel(self.level)
        self.handlers = [console_handler]

        if filename:
            file_handler = logging.FileHandler(filename)
            fmt = "%(levelname)s:%(asctime)s:%(name)s:%(message)s"
            file_handler.setFormatter(logging.Formatter(fmt))
            file_handler.setLevel(self.level)
            self.handlers.append(file_handler)

    def __enter__(self):
        # configure logger with given parameters
        self.logger.setLevel(self.level)
        for handler in self.handlers:
            self.logger.addHandler(handler)

    def __exit__(self, *exc_info):
        # revert back to old logger configuration
        self.logger.setLevel(self.old_level)
        for handler in self.handlers:
            self.logger.removeHandler(handler)



import pytest
from typing import Any

from lokii.util.module_file_loader import ModuleFileLoader


@pytest.fixture
def mock_source_file_loader(mocker):
    def patch(exists: bool = False, code: Any = "") -> None:
        mocker.patch("os.path.exists", return_value=exists)
        mocker.patch("inspect.getsource", return_value="string code content")
        mocker.patch(
            "importlib.machinery.SourceFileLoader.get_code", lambda *args: code
        ),

    return patch


def test_load_raise_error_if_file_not_found(mock_source_file_loader):
    mock_source_file_loader(False)
    loader = ModuleFileLoader("/test/path/any.py")
    with pytest.raises(FileNotFoundError):
        loader.load()


def test_load_should_exec_loaded_module(mock_source_file_loader):
    code = 'conf = {"test_env": {}}'
    mock_source_file_loader(True, compile(code, "/test/path/any.py", "exec"))
    loader = ModuleFileLoader("/test/path/any.py")
    loader.load()
    assert loader.module.conf == {"test_env": {}}



import logging
import os
import shutil
import uuid

from lokii.config import CONFIG
from lokii.storage.data_storage import DataStorage
from lokii.storage.batch_iterator import BatchIterator
from lokii.model.node_module import GenNodeModule
from lokii.parse.node_parser import NodeParser
from lokii.parse.group_parser import GroupParser
from lokii.util.perf_timer_context import PerfTimerContext
from lokii.util.graph_analyzer import GraphAnalyzer
from lokii.exec.node_executor import NodeExecutor

logger = logging.getLogger("lokii")


class Lokii:
    def __init__(self, source_folder: str):
        """
        Generates massive amount of relational mock data.

        :param source_folder: path of root folder that contains schema and table definitions
        """
        self.__source_folder = source_folder
        self.__gen_id = str(uuid.uuid4())

        Lokii.setup_env()
        self.__data_storage = DataStorage()
        self.__node_parser = NodeParser(self.__source_folder)
        self.__group_parser = GroupParser(self.__source_folder)

    def generate(self, export: bool = False, purge: bool = False):
        with PerfTimerContext() as t:
            nodes = self.__node_parser.parse()

            # create dependency map from node source queries
            dep_map = list(self.order_nodes(nodes))
            analyzer = GraphAnalyzer(dep_map)
            exec_order = analyzer.execution_order()

            total_target_count, total_item_count = 0, 0
            for name in exec_order:
                run = nodes[name]
                dep_keys = analyzer.dependencies(name)
                if self.is_node_valid(run, dep_keys):
                    logger.info("%s not changed. Using existing dataset." % name)
                    continue

                # generate dataset
                target_count, item_count, file_paths = self.generate_node(run)
                total_target_count += target_count
                total_item_count += item_count

                # save generation metadata in database
                self.__data_storage.save(self.__gen_id, run.name, run.version)
                # insert generated data in database
                self.__data_storage.insert(run.name, file_paths)

        logger.info("Generation completed!")
        logger.info("Total target item count: {:,}".format(total_target_count))
        logger.info("Generated {:,} items in {}".format(total_item_count, t))

        if export:
            self.export(nodes)
        Lokii.clean_env(purge)

    def generate_node(self, node: GenNodeModule) -> (int, int, list[str]):
        with PerfTimerContext() as t:
            logger = logging.getLogger(node.name)

            executor = NodeExecutor(node, self.__data_storage)
            target_count = executor.prepare_node()

            logger.info("Generation started for target {:,} items".format(target_count))
            generated_file_paths = executor.exec_node()
            item_count = executor.g_count

        logger.info("{:,} items generated in {}".format(item_count, t))
        return target_count, item_count, generated_file_paths

    def export(self, nodes):
        exported = {n: False for n in nodes.keys()}
        with PerfTimerContext() as t:
            groups = self.__group_parser.parse()
            group_nodes = {
                g: [n.name for n in nodes.values() if g in n.groups]
                for g in groups.keys()
            }
            # create dependency map from node source queries
            dep_map = [
                (g.name, [gr for gr in g.groups if gr in groups])
                for g in groups.values()
            ]
            analyzer = GraphAnalyzer(dep_map)
            exec_order = analyzer.execution_order()

            # start from root and exec found `before` functions
            for name in exec_order:
                logger = logging.getLogger(name)
                if groups[name].before:
                    logger.info("Executing before()")
                    groups[name].before({"nodes": group_nodes[name]})
                else:
                    logger.info("before() not executed.")

            # start from leafs and exec found `export` functions
            for name in exec_order[::-1]:
                logger = logging.getLogger(name)
                if groups[name].export:
                    exports = [
                        k
                        for k in nodes.keys()
                        if not exported[k] and k in group_nodes[name]
                    ]
                    for n in exports:
                        logger.info("Executing export() for node %s" % n)
                        cols = self.__data_storage.cols(n)
                        # execute export function for each node in this group
                        iterator = BatchIterator(self.__data_storage, n)
                        groups[name].export(
                            {"name": n, "cols": cols, "batches": iterator}
                        )
                        exported[n] = True
                else:
                    logger.info("export() not executed.")

            # start from leafs and exec found `after` functions
            for name in exec_order[::-1]:
                logger = logging.getLogger(name)
                if groups[name].after:
                    logger.info("Executing after()")
                    groups[name].after({"nodes": group_nodes[name]})
                else:
                    logger.info("after() not executed.")
        export_count = len([k for k in exported.keys() if exported[k]])
        logger.info("{:,} nodes exported in {}".format(export_count, t))

    def is_node_valid(self, node: GenNodeModule, dep_keys: list[str]):
        metadata = self.__data_storage.meta(dep_keys)
        curr = [m for m in metadata if m["name"] == node.name]
        if len(curr) == 0:
            # no dataset generated before with this run key
            return False  # must generate
        if curr[0]["version"] != node.version:
            # code version is different from previous run
            return False  # must regenerate
        for dep_meta in metadata:
            if dep_meta["gen_id"] == self.__gen_id:
                # dependent dataset changed in current generation
                return False  # must regenerate
        # no code changes, no dependencies changed
        return True  # dataset is valid, do not regenerate

    def order_nodes(self, nodes):
        for n in nodes.values():
            # get dependencies in node source query
            deps = self.__data_storage.deps(n.source)
            # check if dependency exists in parsed nodes
            for dep in deps:
                if dep not in nodes:
                    raise AssertionError(
                        "Dependency `%s` in node source query is not found!\n"
                        "Node: %s" % (dep, n.name)
                    )
            logger.debug("Dependencies found for `%s`: %s" % (n.name, deps))
            yield n.name, deps

    @staticmethod
    def setup_env():
        Lokii.clean_env(False)
        if not os.path.exists(CONFIG.temp.data_path):
            os.makedirs(CONFIG.temp.data_path)

    @staticmethod
    def clean_env(force: bool = False):
        if force and os.path.exists(CONFIG.temp.dir_path):
            shutil.rmtree(CONFIG.temp.dir_path)
        elif os.path.exists(CONFIG.temp.data_path):
            shutil.rmtree(CONFIG.temp.data_path)



from typing import Callable

ExportFunc = Callable[[dict], None]


class GroupModule:
    """
    Module specification that defines node group export configuration.

    :var before: Called once before export function
    :type before: ExportFunc
    :var after: Called once after export function
    :type after: ExportFunc
    :var export: Called for every node in this group
    :type export: ExportFunc
    :var name: Name of the group
    :type name: str
    :var groups: Groups that the node belongs
    :type groups: list[str]
    """

    def __init__(self, name, groups=None):
        """
        Initialize generation node group module.
        :param name: name of the group
        :type name: str
        :param groups: export groups of the node
        :type groups: list[str]
        """
        self.name = name
        self.groups = groups or []
        self.before = None
        self.after = None
        self.export = None



from logging import Formatter, LogRecord
from os import path

from lokii.logger.color import default_formats


class MultiFormatter(Formatter):
    """Format log messages differently for each log level"""

    def __init__(self, formats: dict[int, str] = None, **kwargs):
        base_format = kwargs.pop("fmt", None)
        super(MultiFormatter, self).__init__(base_format, **kwargs)
        fmts = formats or default_formats
        self.formatters = {lvl: Formatter(fmt, **kwargs) for lvl, fmt in fmts.items()}

    def format(self, r: LogRecord):
        fmt = self.formatters.get(r.levelno)
        if hasattr(r, "at"):
            file_path = "'file:%s%s%s'" % (path.sep, path.sep, path.abspath(r.at))
            r.msg = "%s at %s" % (r.msg, file_path)
            delattr(r, "at")
        return super(MultiFormatter, self).format(r) if fmt is None else fmt.format(r)



import logging
from os import path
from glob import glob

from lokii.config import CONFIG
from lokii.model.group_module import GroupModule
from lokii.util.module_file_loader import ModuleFileLoader
from lokii.util.perf_timer_context import PerfTimerContext
from lokii.parse.base_parser import BaseParser

logger = logging.getLogger("lokii.group_parser")
GROUP_RESOLVER = "**/*%s" % CONFIG.gen.group_ext


class GroupParser(BaseParser):
    """
    Reads and validates group export configurations from filesystem structure.

    :var groups: parsed group modules
    :type groups: dict[str, GroupModule]
    """

    def __init__(self, source_folder):
        """
        Initializes parser.
        :param source_folder: root path of the project
        :type source_folder: str
        """
        self.root = source_folder
        self.groups = {}

    def parse(self):
        """
        :return: parsed and validated node run flat map
        :rtype: dict[str, GroupModule]
        """
        with PerfTimerContext() as t:
            groups = list(self.__parse_groups())
            self.groups = {n.name: n for n in groups}

        group_count = len(groups)
        if group_count == 0:
            logger.warning("No group conf file found", extra={"at": self.root})
        else:
            logger.info("%d group definitions parsed in %s" % (group_count, t))
        return self.groups

    def __parse_groups(self):
        """
        Finds all the groups matching the group file pattern. Loads group modules and ensures
        group module configuration is valid.

        :return: parsed and validated group modules
        :rtype: list[GroupModule]
        """
        glob_path = path.join(self.root, GROUP_RESOLVER)
        file_paths = [f for f in glob(glob_path, recursive=True)]

        for fp in file_paths:
            loader = ModuleFileLoader(fp)
            loader.load()
            mod = loader.module

            # extract group name from dir name
            m_name = path.dirname(fp).split(path.sep)[-1]
            # extract groups from file path
            m_groups = path.relpath(path.dirname(fp), self.root).split(path.sep)[:-1]
            parsed = GroupModule(m_name, m_groups)
            logger.debug("Found valid group `%s`", m_name, extra={"at": fp})

            # ensure group configuration is valid
            if self.attr(mod, "before"):
                self.func(mod.before, "`before` must be function at %s" % fp)
                self.sig(mod.before, 1, "`before` accepts only one param at %s" % fp)
                parsed.before = mod.before
            if self.attr(mod, "export"):
                self.func(mod.export, "`export` must be function at %s" % fp)
                self.sig(mod.export, 1, "`export` accepts only one param at %s" % fp)
                parsed.export = mod.export
            if self.attr(mod, "after"):
                self.func(mod.after, "`after` must be function at %s" % fp)
                self.sig(mod.after, 1, "`after` accepts only one param at %s" % fp)
                parsed.after = mod.after
            yield parsed



from typing import Callable

GenItemFunc = Callable[[dict], dict]


class GenNodeModule:
    """
    Module specification that defines node generation configuration.

    :var source: source query to create items
    :type source: str
    :var item: item generation function
    :type item: GenFunc
    :var name: Name of the node
    :type name: str
    :var version: Code version of the node
    :type version: str
    :var groups: Groups that the node belongs
    :type groups: list[str]
    """

    def __init__(self, source, item, name=None, version=None, groups=None):
        """
        Initialize generation node module.
        :param source: source query to create items
        :type source: str
        :param item: item generation function
        :type item: GenFunc
        :param name: name of the node
        :type name: str
        :param groups: export groups of the node
        :type groups: list[str]
        """
        self.source = source
        self.item = item
        self.name = name
        self.version = version
        self.groups = groups or []



import logging

_ansi_colors = {
    "black": 30,
    "red": 31,
    "green": 32,
    "yellow": 33,
    "blue": 34,
    "magenta": 35,
    "cyan": 36,
    "white": 37,
    "reset": 39,
    "bright_black": 90,
    "bright_red": 91,
    "bright_green": 92,
    "bright_yellow": 93,
    "bright_blue": 94,
    "bright_magenta": 95,
    "bright_cyan": 96,
    "bright_white": 97,
}
_ansi_reset_all = "\033[0m"


def style(
    text: str, fg: str = None, bg: str = None, bold: bool = False, reset: bool = True
):
    bits = []
    bit_temp = "\033[%dm"
    if fg:
        bits.append(bit_temp % _ansi_colors[fg])
    if bg:
        bits.append(bit_temp % (_ansi_colors[bg] + 10))
    bits.append(bit_temp % (1 if bold else 22))
    bits.append(text)
    if reset:
        bits.append(_ansi_reset_all)
    return "".join(bits)


default_formats = {
    logging.DEBUG: style("DEBUG | %(name)s > %(message)s", fg="bright_green"),
    logging.INFO: style("INFO  | %(name)s > %(message)s", fg="cyan"),
    logging.WARNING: style("WARN  | %(name)s > %(message)s", fg="yellow"),
    logging.ERROR: style("ERROR | %(name)s > %(message)s", fg="red"),
    logging.CRITICAL: style("FATAL", fg="bright_white", bg="red", bold=True)
    + style(" | %(message)s", fg="red", bold=True),
}



from unittest.mock import Mock

import pytest
import os

from lokii.storage.data_storage import DataStorage
from lokii.exec.node_executor import NodeExecutor, _exec_chunk
from lokii.model.node_module import GenNodeModule


@pytest.mark.usefixtures("setup_test_env")
def test__exec_chunk_should_call_func_for_list_length():
    func = Mock()
    _exec_chunk(func, [{"data": f} for f in range(100)])
    assert func.call_count == 100


@pytest.mark.usefixtures("setup_test_env")
def test_prepare_node_should_return_total_target_count_for_query():
    data_storage = DataStorage()
    node = GenNodeModule("SELECT 1", lambda x: x, name="n1")
    executor = NodeExecutor(node, data_storage)
    assert executor.prepare_node() == 1


@pytest.mark.usefixtures("setup_test_env")
def test_exec_node_should_return_generated_list_of_files():
    data_storage = DataStorage()
    node = GenNodeModule("SELECT 1", lambda x: x, name="n1")
    executor = NodeExecutor(node, data_storage)
    executor.prepare_node()
    files = executor.exec_node()
    assert len(files) == 1
    assert os.path.exists(files[0])



import logging
from os import path
from glob import glob

from lokii.config import CONFIG
from lokii.model.node_module import GenNodeModule
from lokii.util.module_file_loader import ModuleFileLoader
from lokii.util.perf_timer_context import PerfTimerContext
from lokii.parse.base_parser import BaseParser

logger = logging.getLogger("lokii.node_parser")
NODE_RESOLVER = "**/*%s" % CONFIG.gen.node_ext


class NodeParser(BaseParser):
    """
    Reads and validates node configurations from filesystem structure.

    :var nodes: parsed generation nodes
    :type nodes: dict[str, GenNodeModule]
    """

    def __init__(self, source_folder):
        """
        Initializes parser.
        :param source_folder: root path of the project
        :type source_folder: str
        """
        self.root = source_folder
        self.nodes = {}

    def parse(self):
        """
        :return: parsed and validated node run flat map
        :rtype: dict[str, GenNodeModule]
        """
        with PerfTimerContext() as t:
            nodes = list(self.__parse_nodes())
            self.nodes = {n.name: n for n in nodes}

        node_count = len(nodes)
        if node_count == 0:
            logger.warning("No generation node file found", extra={"at": self.root})
        else:
            logger.info("%d generation node parsed in %s" % (node_count, t))
        return self.nodes

    def __parse_nodes(self):
        """
        Finds all the nodes matching the node file pattern. Loads node modules and ensures
        node module configuration is valid.

        :return: parsed and validated node modules
        :rtype: list[GenNodeModule]
        """
        glob_path = path.join(self.root, NODE_RESOLVER)
        file_paths = [f for f in glob(glob_path, recursive=True)]

        for fp in file_paths:
            loader = ModuleFileLoader(fp)
            loader.load()
            mod = loader.module

            # extract node name from filename
            m_name = path.basename(fp).replace(CONFIG.gen.node_ext, "")
            # extract groups from file path
            m_groups = path.relpath(fp, self.root).split(path.sep)[:-1]
            m_version = loader.version

            # ensure provided module is valid
            self.attr(mod, "source", "`source` not found at %s" % fp)
            self.inst(mod.source, str, "`source` must be str at %s" % fp)
            self.attr(mod, "item", "`item` not found at %s" % fp)
            self.func(mod.item, "`item` must be function at %s" % fp)
            self.sig(mod.item, 1, "`item` accepts only one param at %s" % fp)
            if self.attr(mod, "name"):
                self.inst(mod.name, str, "`name` must be str at %s" % fp)
                m_name = mod.name

            parsed = GenNodeModule(mod.source, mod.item, m_name, m_version, m_groups)
            logger.debug("Found valid node `%s`", m_name, extra={"at": fp})
            yield parsed



import os
import shutil
import pytest
from io import StringIO

from config import CONFIG
from lokii import Lokii
from lokii.cli import exec_cmd, LOKII_ASCII


def test_exec_cmd_should_print_help(mocker):
    mock = mocker.patch("sys.stdout", new=StringIO())
    with pytest.raises(SystemExit):
        exec_cmd("lokii -f tests -h")
    assert LOKII_ASCII in mock.getvalue()
    Lokii.clean_env(force=True)


def test_exec_cmd_should_create_log_file():
    log_file_path = "test.log"
    exec_cmd("lokii -f tests -l %s" % log_file_path)
    assert os.path.exists(log_file_path)
    os.remove(log_file_path)


def test_exec_cmd_should_obey_log_level(caplog):
    exec_cmd("lokii -f tests -q")
    assert len(caplog.text) == 0
    exec_cmd("lokii -f tests -v")
    assert len(caplog.text) > 0


@pytest.mark.parametrize(
    "glob_files, load_modules",
    [(["t1.node.py"], [{"source": "INVALID SQL", "item": lambda x: x}])],
    indirect=True,
)
def test_exec_cmd_should_log_errors_with_critical_level(
    glob_files, load_modules, caplog
):
    exec_cmd("lokii -f tests")
    assert "CRITICAL" in caplog.text


@pytest.mark.parametrize("glob_files, load_modules", [([], [])], indirect=True)
def test_exec_cmd_should_clear_db_if_purge(glob_files, load_modules):
    exec_cmd("lokii -f tests -p")
    assert not os.path.exists(CONFIG.temp.db_path)


@pytest.mark.parametrize(
    "glob_files, load_modules",
    [(["t1.node.py"], [{"source": "SELECT 1", "item": lambda x: x}])],
    indirect=True,
)
def test_exec_cmd_should_cache_consequent_runs(glob_files, load_modules, caplog):
    exec_cmd("lokii -f tests")
    assert os.path.exists(CONFIG.temp.db_path)
    caplog.clear()

    exec_cmd("lokii -p -f tests")
    assert "not changed. Using existing dataset." in caplog.text


gen_mod = {"source": "SELECT 1", "item": lambda x: x, "version": "v1"}


@pytest.mark.usefixtures("glob_files", "load_modules")
@pytest.mark.parametrize(
    "glob_files, load_modules", [(["t1.node.py"], [gen_mod])], indirect=True
)
def test_exec_cmd_should_regenerate_on_code_change(caplog):
    # start generation for code version v1
    gen_mod["version"] = "v1"
    exec_cmd("lokii -f tests")
    assert os.path.exists(CONFIG.temp.db_path)
    caplog.clear()

    # start generation for code version v2
    gen_mod["version"] = "v2"
    exec_cmd("lokii -f tests -p")
    # should regenerate and not use cache
    assert "not changed. Using existing dataset." not in caplog.text


gen_mod1, gen_mod2 = (
    {"source": "SELECT * from n2", "item": lambda x: x, "name": "n1"},
    {"source": "SELECT 1", "item": lambda x: x, "name": "n2", "version": "v1"},
)


@pytest.mark.usefixtures("glob_files", "load_modules")
@pytest.mark.parametrize(
    "glob_files, load_modules",
    [(["t1.node.py", "t2.node.py"], [gen_mod1, gen_mod2])],
    indirect=True,
)
def test_exec_cmd_should_regenerate_on_dependency_change(caplog):
    # start generation for dependency code version v1
    gen_mod2["version"] = "v1"
    exec_cmd("lokii  -f tests")
    assert os.path.exists(CONFIG.temp.db_path)
    caplog.clear()

    # start generation for dependency code version v2
    gen_mod2["version"] = "v2"
    exec_cmd("lokii -f tests -p")
    # should regenerate and not use cache because of dependency change
    assert "not changed. Using existing dataset." not in caplog.text



import logging
import os.path
from functools import partial

import duckdb
from typing import TypedDict, Callable

from lokii.config import CONFIG

NodeMetadata = TypedDict("NodeMetadata", {"name": str, "version": str, "gen_id": str})


class DataStorage:
    def __init__(self):
        """
        Temporary filesystem storage implementation for storing data generated between batches.
        It only stores data temporary and deletes all files after
        """
        self.connect = partial(duckdb.connect, database=CONFIG.temp.db_path)
        with self.connect() as conn:
            # create node meta table to store run generation information
            q = (
                "CREATE TABLE IF NOT EXISTS __meta"
                "(name TEXT, version TEXT, gen_id TEXT, PRIMARY KEY(name));"
            )
            conn.execute(q).fetchall()

    def deps(self, query: str) -> list:
        names = duckdb.get_table_names(query)
        return list(names)

    def count(self, query: str) -> int:
        q = "WITH _t AS (%s) SELECT COUNT() FROM _t;" % query
        try:
            with self.connect() as conn:
                (count,) = conn.execute(q).fetchone()
                return count
        except duckdb.Error as err:
            logging.error("Error occurred while executing count query:\n\n%s\n" % q)
            raise err

    def exec(self, query: str, index: int, size: int) -> list[dict]:
        args = (query, size, index * size)
        q = "WITH _t AS (%s) SELECT * FROM _t LIMIT %d OFFSET %d;" % args
        try:
            with self.connect() as conn:
                data = conn.execute(q).fetch_df()
                return data.to_dict("records")
        except duckdb.Error as err:
            logging.error("Error occurred while executing source query:\n\n%s\n" % q)
            raise err

    def save(self, gen_id: str, name: str, version: str) -> None:
        """
        Generation id and node version will be stored in a meta table that can be used to check
        if gen run data is valid for consecutive runs.

        :param gen_id: identification of the generation process
        :param name: identification of the node
        :param version: the code version of the module
        """
        with self.connect() as conn:
            q = """
            INSERT OR REPLACE INTO __meta(name, version, gen_id)
            VALUES ('%s', '%s', '%s');
            """
            conn.execute(q % (name, version, gen_id)).fetchall()

    def meta(self, names: list[str]) -> list[NodeMetadata]:
        """
        Fetches generation metadata information about given runs.
        :param names: list of node names
        :return: list of node meta dict
        """
        with self.connect() as conn:
            keys = ",".join(["'%s'" % n for n in names])
            q = "SELECT name, version, gen_id FROM __meta WHERE name IN (%s);"
            data = conn.execute(q % keys).fetch_df()
            return data.to_dict("records")

    def cols(self, name) -> list[str]:
        """
        Fetches generated column name list for given node.
        :param name: name of the node
        :type name: str
        :return: list of no columns
        """
        with self.connect() as conn:
            q = "DESCRIBE %s;"
            data = conn.execute(q % name).fetchall()
            return [col[0] for col in data]

    def insert(self, name: str, files: list[str]) -> None:
        """
        Creates a table for given node name in local relational database. If there is a table
        with the same node name it will drop all data and create a fresh one. Given files will
        be concatenated and inserted in the fresh table.

        :param name: node name of the module
        :param files: list of generated file paths
        """
        with self.connect() as conn:
            assert "." not in name, "Node names can not contain dot(.) = %s" % name

            # concatenate and insert file contents in a fresh table
            q = "CREATE OR REPLACE TABLE %s AS SELECT * FROM read_json_auto(%s);"
            q = q % (name, files)
            if len(files) == 0:
                q = "CREATE OR REPLACE TABLE %s(id INTEGER);"
                q = q % name
            conn.execute(q).fetchall()

    def export(self, out_path: str, fmt: str) -> None:
        """
        Exports all generated tables to given file format.

        :param out_path: folder path of the exported files
        :param fmt: file format of the exported files
        """
        with self.connect() as conn:
            tables = conn.execute("SHOW TABLES;").fetchall()
            # exclude meta table from export list
            tables = [f for (f,) in tables if "__meta" != f]

            # create out folder if not exists
            if not os.path.exists(out_path):
                os.makedirs(out_path)

            for t in tables:
                path = os.path.join(out_path, ".".join([t, fmt]))
                # export to output path
                conn.execute("COPY %s TO '%s'" % (t, path)).fetchone()



import pytest

from lokii.parse.group_parser import GroupParser

pytestmark = [pytest.mark.usefixtures("glob_files", "load_modules")]


@pytest.mark.parametrize("glob_files, load_modules", [([], [])], indirect=True)
def test_parse_should_log_warning_if_no_modules_found(glob_files, load_modules, caplog):
    GroupParser("/test/proj").parse()
    assert "No group conf file found" in caplog.text


@pytest.mark.parametrize("load_modules", [[{}]], indirect=True)
@pytest.mark.parametrize(
    "glob_files, expect",
    [(["/test/proj/test.group.py"], "proj"), (["/test/proj/g1/test.group.py"], "g1")],
    indirect=["glob_files"],
)
def test_parse_should_use_folder_name_as_group_identifier(
    glob_files, load_modules, expect
):
    parsed = GroupParser("/test/proj").parse()
    assert expect in parsed
    assert expect == parsed[expect].name


@pytest.mark.parametrize(
    "glob_files, load_modules, expect",
    [
        (
            ["/test/path/g1/t1.group.py", "/test/path/g1/g2/t2.group.py"],
            [{"name": "g1"}, {"name": "g2"}],
            [[], ["g1"]],
        )
    ],
    indirect=["glob_files", "load_modules"],
)
def test_parse_should_use_file_path_to_extract_groups(glob_files, load_modules, expect):
    parser = GroupParser("/test/path")
    parser.parse()
    for i, module_path in enumerate(glob_files):
        node_name = load_modules[i]["name"]
        assert expect[i] == parser.groups[node_name].groups


@pytest.mark.parametrize("glob_files", [["/test/path/g1/g1.group.py"]], indirect=True)
@pytest.mark.parametrize(
    "load_modules, expect",
    [
        ([{"before": ""}], "`before` must be function"),
        ([{"before": lambda x, y: x}], "`before` accepts only one param"),
        ([{"export": ""}], "`export` must be function"),
        ([{"export": lambda x, y: x}], "`export` accepts only one param"),
        ([{"after": ""}], "`after` must be function"),
        ([{"after": lambda x, y: x}], "`after` accepts only one param"),
    ],
    indirect=["load_modules"],
)
def test_parse_raise_error_if_module_not_valid(glob_files, load_modules, expect):
    with pytest.raises(AssertionError) as err:
        GroupParser("/test/proj").parse()
    assert expect in str(err.value)
